-- |
-- Module	: NMUTool.Packages
-- Copyright	: 2009 Kari Pahula <kaol@debian.org>
-- License	: GPLv2 or later

module NMUTool.Packages (loadPackages) where

import Text.ParserCombinators.Parsec
import Control.Monad (guard,mplus)
import Data.Maybe (catMaybes, fromMaybe, maybeToList)
import Data.Char (isAlpha)
import Data.List (nub,find)
import System.IO (hGetContents,Handle)
import NMUTool.Version
import NMUTool.Types
import Debug.Trace

loadPackages :: Name -> Handle -> IO [Source]
loadPackages depmatch h = fmap ((transClosure depmatch) . mergeBin . collat . lines) $ hGetContents h

-- transitive closure of requested package's rdeps
transClosure :: Name -> [Source] -> [Source]
transClosure dep allpkgs = twoEq $ iterate closure depPkg
    where
      depPkg = filter ((dep `elem`) . sdeps) allpkgs
      twoEq (a:r@(b:_)) = if a == b then a else twoEq r
      closure deps = nub $ deps ++ (filter (inDep deps) allpkgs)
      inDep deps pkg = any (`elem` (sdeps pkg)) (concatMap sbins deps)

depParser :: Parser [Name]
depParser = do
  skipMany (space <|> char ':')
  fmap catMaybes (sepBy1 (do bin <- many1 $ alphaNum <|> oneOf "-."
                             skipMany space
                             ((do char '('
                                  dep <- manyTill anyChar $ try $ char ')'
                                  return $ do guard $ (take 2 dep) == "<<"
                                              return bin)
                              <|> (return Nothing)))
                  ((char ',') >> skipMany space))

mergeBin :: [Source] -> [Source]
mergeBin srcs =
    foldr (\(n,(v,_,_)) ss ->
               (case lookup n ss of
                  Nothing -> ((n,(v,concatMap (\(_,(_,bs',_)) -> bs') $ filter ((== n) . fst) srcs,[])):)
                  _ -> id) ss
          ) [] srcs

collat :: [String] -> [Source]
collat =
    (\(_,_,c) -> c) . foldr collat' ([],[],[])
    where
      collat' "" (_,m,c) =
          ([],[],fromMaybe c $ do
             [pkg] <- lookup "Package" m
             [src] <- lookup "Source" m `mplus` return [pkg]
             [ver] <- (fmap . fmap) sourceVersion $ lookup "Version" m
             prv <- lookup "Provides" m `mplus` return []
             dep <- lookup "Depends" m
             return $ (takeWhile (/= ' ') src,(ver,[Binary { binary = pkg, devdeps = dep, provides = prv }],[])):c)
      collat' l (ds,m,c) =
          let (field,r) = span isAlpha l in
          case field of
            "" -> ((r:ds),m,c)
            _ -> ([],if (field `elem` ["Depends", "Provides"])
                     then (field, case parse depParser "" r of Right x -> x):m
                     else (if field `elem` ["Source", "Package", "Version"]
                           then (field,[(dropWhile (`elem` " :") r)
                                        ++ (foldr (++) "" ds)]):m
                           else m),c)


{-
test = [("ghc",((Version ""),[Binary "ghc" [] []],[]))
       ,("mtl",((Version ""),[Binary "mtl-dev" ["ghc"] []],[]))
       ,("parsec",((Version ""),[Binary "parsec-dev" ["mtl-dev","ghc"] []],[]))
       ,("network",((Version ""),[Binary "network-dev" ["parsec-dev","ghc"] []],[]))
       ,("http",((Version ""),[Binary "http-dev" ["network-dev","ghc"] []],[]))
       ,("ftp",((Version ""),[Binary "ftp-dev" ["network-dev","ghc"] []],[]))
       ]       
-}
